from django.db.models import Q
from django.utils import timezone
from rest_framework import status, viewsets
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from dj_notif.managers import NotificationManager, BroadcastNotificationManager


class BaseNotificationViewSet(viewsets.ViewSet):
    permission_classes = [IsAuthenticated]

    @staticmethod
    def _parse_flags(request):
        unread_only = request.query_params.get('unread_only', '').lower() == 'true'
        categorized = request.query_params.get('categorized', '').lower() == 'true'
        page = request.query_params.get('page', 1)
        return unread_only, categorized, page


class UserNotificationViewSet(BaseNotificationViewSet):

    def list(self, request):
        unread_only, categorized, _ = self._parse_flags(request)
        manager = NotificationManager(user=request.user)

        if categorized:
            serialized = manager.get_categorized_notifications_serialized(unread_only=unread_only)
        else:
            serialized = manager.get_notifications_serialized(unread_only=unread_only)

        return Response(serialized)

    @action(detail=False, methods=['get'])
    def categories(self, request):
        manager = NotificationManager(user=request.user)
        result = manager.get_categories_unread_count()
        return Response(result)

    @action(detail=False, methods=['post'])
    def mark_all_as_read(self, request):
        manager = NotificationManager(user=request.user)
        total_updated = sum(
            model.objects.filter(user=request.user, read_at=None)
            .update(read_at=timezone.localtime())
            for model in manager.get_notification_models().values()
        )
        return Response({"detail": f"{total_updated} notifications marked as read."})

    @action(detail=False, methods=['get'], url_path='category/(?P<category>[^/.]+)')
    def category_notifications(self, request, category=None):
        unread_only, _, _ = self._parse_flags(request)
        manager = NotificationManager(user=request.user)
        categories = manager.get_notification_categories()

        if category not in categories:
            return Response({'detail': 'Category not found'}, status=status.HTTP_404_NOT_FOUND)

        serialized = manager.get_category_notifications_serialized(category, unread_only=unread_only)
        return Response(serialized)

    @action(detail=False, methods=['post'], url_path='category/(?P<category>[^/.]+)/mark_all_read')
    def mark_category_read(self, request, category=None):
        manager = NotificationManager(user=request.user)
        if category not in manager.get_notification_categories():
            return Response({'detail': 'Category not found'}, status=status.HTTP_404_NOT_FOUND)

        model = manager.get_notification_models()[category]
        updated = model.objects.filter(user=request.user, read_at=None).update(read_at=timezone.localtime())

        return Response({"detail": f"{updated} notifications marked as read."})

    @action(detail=False, methods=['post'], url_path='category/(?P<category>[^/.]+)/mark_read/(?P<notification_id>\d+)')
    def mark_single_notification_read(self, request, category=None, notification_id=None):
        manager = NotificationManager(user=request.user)
        try:
            notif = manager.get_notification_models()[category].objects.get(id=notification_id, read_at=None)
            notif.read_at = timezone.localtime()
            notif.save(update_fields=['read_at'])
            return Response({'detail': f"Notification {notification_id} marked as read"})
        except Exception as e:
            return Response({'detail': str(e)}, status=status.HTTP_404_NOT_FOUND)


class BroadcastNotificationViewSet(BaseNotificationViewSet):

    def list(self, request):
        unread_only, categorized, _ = self._parse_flags(request)
        manager = BroadcastNotificationManager()

        if categorized:
            serialized = manager.get_categorized_notifications_serialized(requested_user=request.user)
        else:
            serialized = manager.get_notifications_serialized(requested_user=request.user)

        return Response(serialized)

    @action(detail=False, methods=['get'], url_path='category/(?P<category>[^/.]+)')
    def get(self, request, category=None):
        manager = BroadcastNotificationManager()
        categories = manager.get_notification_categories()

        if category not in categories:
            return Response({'detail': 'Category not found'}, status=status.HTTP_404_NOT_FOUND)

        serialized = manager.get_category_notifications_serialized(category, requested_user=request.user)
        return Response(serialized)

    @action(detail=False, methods=['post'], url_path='category/(?P<category>[^/.]+)/mark_read/(?P<notification_id>\d+)')
    def mark_single_notification_read(self, request, category=None, notification_id=None):
        manager = BroadcastNotificationManager()
        user_id = str(request.user.id)
        try:
            notif = manager.get_notification_models()[category].objects.filter(
                Q(**{f"user_read_at_receipts__{'16'}": False}) |
                ~Q(**{f"user_read_at_receipts__has_key": "16"}),
                id=notification_id
            ).get()

            formatted_time = timezone.localtime().strftime('%Y-%m-%d %H:%M:%S')
            notif.user_read_at_receipts[user_id] = formatted_time
            notif.save(update_fields=['user_read_at_receipts'])
            return Response({'detail': f"Notification {notification_id} marked as read"})
        except Exception as e:
            return Response({'detail': str(e)}, status=status.HTTP_404_NOT_FOUND)



import logging
import json
from dj_notif.base.consumers import BaseConsumer
from dj_notif.utils.main import MessageResult
from dj_notif.utils.async_funcs import get_notifications, mark_as_read, mark_all_as_read

logger = logging.getLogger('notification')


class NotificationConsumer(BaseConsumer):
    NotificationTypesActionsMap = {
        "notifications": {
            1: "get",
            2: "mark_as_read",
            3: "mark_all_as_read"
        }
    }

    # def get_serializer_class(self, action_type, action_name):
    #     from serializers import BaseNotificationConsumerSerializer
    #     exists_serializers = BaseNotificationConsumerSerializer.__subclasses__()

    async def handle_action(self, action_type, action_name, payload):
        match (action_type, action_name):

            # TODO future update: serializer for ws notification manager
            # case ('notifications', 'get'):
            #     project_id = payload.get("project_id")  # NOT REQUIRED, FILTER
            #     task_id = payload.get("task_id")  # NOT REQUIRED, FILTER
            #     result = await get_notifications(self.user, project_id, task_id)
            #     await self.send_response(action_type, action_name, result)
            #
            # case ('notifications', 'mark_as_read'):
            #     notification_id = payload.get('notification_id')  # REQUIRED
            #     result = await mark_as_read(self.user, notification_id)
            #     await self.send_response(action_type, action_name, result)
            #
            # case ('notifications', 'mark_all_as_read'):
            #     project_id = payload.get("project_id")  # NOT REQUIRED, FILTER
            #     task_id = payload.get("task_id")  # NOT REQUIRED, FILTER
            #     result = await mark_all_as_read(self.user, project_id, task_id)
            #     await self.send_response(action_type, action_name, result)

            case _:
                logger.error(f"Unknown action_type: {action_type} or action_name: {action_name}")
                result = MessageResult("Unknown action", False, 404)
                await self.send_response(action_type, action_name, result)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logger

    async def notifications(self, event):
        await self.send(text_data=json.dumps(event))



import json

from channels.generic.websocket import AsyncWebsocketConsumer


class BaseConsumer(AsyncWebsocketConsumer):
    def __init__(self, *args, **kwargs):
        self.group_name = None
        self.user = None
        self.logger = None
        super().__init__(*args, **kwargs)

    async def connect(self, **kwargs):
        self.user = self.scope.get('user')

        if self.user and self.channel_layer:

            self.group_name = f"notif_user_{self.user.id}"

            await self.channel_layer.group_add(self.group_name, self.channel_name)
            await self.channel_layer.group_add('broadcast', self.channel_name)
            await self.accept()
        else:
            await self.close()

    async def disconnect(self, close_code):
        if self.group_name and self.channel_layer:
            await self.channel_layer.group_discard(self.group_name, self.channel_name)

    async def receive(self, text_data=None, bytes_data=None):
        try:
            data = json.loads(text_data)
            action_type = data.get('type')
            action_name = data.get('action')
            payload = data.get("payload", {})

            self.logger.debug(f"Received action: {action_type} - {action_name} with payload: {payload}")

            # Handle the action based on its type using match-case
            await self.handle_action(action_type, action_name, payload)

        except json.JSONDecodeError:
            self.logger.error("Invalid JSON format received.")
        except Exception as e:
            self.logger.exception(f"Error processing WebSocket message: {str(e)}")

    async def handle_action(self, action_type, action_name, payload):
        ...

    async def send_response(self, action_type, action_name, result):
        message_type = "result" if result.success else "error"

        await self.channel_layer.group_send(
            self.group_name,
            {
                'type': action_type,
                'action': action_name,
                'message': f'Action received: {action_name}',
                'payload': {
                    "status_code": result.status,
                    "success": result.success,
                    message_type: result.__str__(),
                    "meta_data": result.meta
                }
            }
        )




from django.urls import path, include
from rest_framework.routers import DefaultRouter

from dj_notif.views import UserNotificationViewSet, BroadcastNotificationViewSet
from django_channels_jwt.views import AsgiValidateTokenView

router = DefaultRouter()
router.register(r'user', UserNotificationViewSet, basename='user-notifications')
router.register(r'broadcast', BroadcastNotificationViewSet, basename='broadcast-notifications')

urlpatterns = [
    path("notifications/ws_auth/", AsgiValidateTokenView.as_view()),  # WS Authentication
    path('notifications', include(router.urls)),
]



from rest_framework import serializers

from dj_notif.base.serializers import BaseSerializer
from dj_notif.utils.main import get_subclasses


class AbstractBaseNotificationSerializer(BaseSerializer):
    user = None
    notification_class_name = None
    category = serializers.SerializerMethodField()

    def __new__(cls, *args, **kwargs):
        if not (_ := getattr(cls, "notification_class_name", None)):
            raise serializers.ValidationError("Define notification_class_name in child class attributes")
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):

        meta = self.Meta
        if not hasattr(meta, "fields"):
            meta.fields = ['id', 'category', 'title', 'description', 'level', 'created_at', 'read_at']

        if extra_fields := getattr(meta, "extra_fields", None):
            meta.fields.extend(extra_fields)

        super().__init__(*args, **kwargs)

    @classmethod
    def get_all_subclasses(cls):
        return get_subclasses(cls, recursive=True)

    @classmethod
    def _get_subclasses(cls) -> dict:
        all_subclasses = cls.get_all_subclasses()
        return {x.notification_class_name: x for x in all_subclasses}

    @classmethod
    def get_subclass(cls, notif_name):
        try:
            return cls._get_subclasses().get(notif_name)
        except Exception as err:
            raise serializers.ValidationError(err)

    def get_category(self, obj):
        return self.notification_class_name


class BaseNotificationSerializer(AbstractBaseNotificationSerializer):
    ...


class BaseBroadcastNotificationSerializer(AbstractBaseNotificationSerializer):
    read_at = serializers.SerializerMethodField()

    def get_read_at(self, obj):
        if request := self.context.get('request', None):
            user = request.user
            return obj.user_read_at_receipts.get(str(user.id), None)

        elif user := self.context.get('user', None):
            return obj.user_read_at_receipts.get(str(user.id), None)

        return None


# TODO future update: serializer for ws notification manager
# class BaseNotificationConsumerSerializer(serializers.Serializer):
#
#     def validate_action(self, action_name: str, action_type: str):
#         ...
#
#     def __init__(self, action_type: str, action_name: str):
#         self.type = action_type
#         self.action = action_name
#         super().__init__()
#
#
# class NotificationGetSerializer(BaseNotificationConsumerSerializer):
#     ...
#
#
# class NotificationMarkAsReadSerializer(BaseNotificationConsumerSerializer):
#     ...
#
#
# class NotificationMarkAllAsReadSerializer(BaseNotificationConsumerSerializer):
#     ...



from django.db import models


class SoftDeleteManager(models.Manager):
    def get_queryset(self):
        # Exclude soft-deleted items by default
        return super().get_queryset().filter(deleted_at__isnull=True)



from django.db import models
from django.utils import timezone

from dj_notif.base.managers import SoftDeleteManager


class BaseModel(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    modified_at = models.DateTimeField(auto_now=True)
    deleted_at = models.DateTimeField(null=True, blank=True)

    class Meta:
        abstract = True  # Set this model as Abstract

    objects = SoftDeleteManager()

    def delete(self, using=None, keep_parents=False):
        self.deleted_at = timezone.localtime()
        self.save(update_fields=['deleted_at'])

    def hard_delete(self, using=None, keep_parents=False):
        return super().delete(using=using, keep_parents=keep_parents)



from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer

from django.db import models
from django.utils import timezone

from dj_notif.base.models import BaseModel
from dj_notif.utils.main import get_subclasses


class MethodNotAllowed(Exception):
    def __init__(self, method_name):
        self.method_name = method_name
        super().__init__(self.method_name)

    def __str__(self):
        return f"Method {self.method_name} not allowed in {self.__class__.__name__}"


class AbstractBaseNotificationModel(BaseModel):
    class NotificationLevel(models.IntegerChoices):
        low = 1,
        medium = 2,
        high = 3

    notification_class_name = None
    requested_user = models.PositiveIntegerField(null=True, blank=True)
    title = models.CharField(max_length=50)
    level = models.IntegerField(choices=NotificationLevel.choices)
    description = models.CharField(max_length=200)

    def __new__(cls, *args, **kwargs):
        if not (_ := getattr(cls, "notification_class_name", None)):
            raise RuntimeError("Define notification_class_name in class attributes")
        return super().__new__(cls)

    class Meta:
        abstract = True

    @classmethod
    def get_all_subclasses(cls):
        return get_subclasses(cls, recursive=True)

    @classmethod
    def get_serializer_class(cls):
        from .serializers import AbstractBaseNotificationSerializer
        return AbstractBaseNotificationSerializer.get_subclass(cls.notification_class_name)

    def dispatch(self, broadcast=False):
        message_prefix = "Broadcast" if broadcast else "New"
        serializer = self.get_serializer_class()(self, excluded_fields=['user'])
        channel_layer = get_channel_layer()
        channel_name = self.get_channel_name()

        async_to_sync(channel_layer.group_send)(
            channel_name,  # This should match the group your WebSocket consumer is listening to
            {
                'type': 'notifications',  # This should match the method name in your WebSocket consumer
                'message': f'{message_prefix} Notification',
                'action': self.__class__.__name__,
                'payload': {
                    'notification': serializer.data
                }
            }
        )

    def get_channel_name(self):
        pass


class BaseNotificationModel(AbstractBaseNotificationModel):
    user = models.PositiveIntegerField()
    read_at = models.DateTimeField(null=True, blank=True, default=None)

    class Meta:
        abstract = True

    @classmethod
    def create_notifications(cls, title: str, description: str, level: int, users: list[int]):
        for user_id in users:
            try:
                cls.objects.create(title=title, description=description, level=level, user=user_id)
            except Exception as err:
                print(err)
        return None

    def get_user_id(self):
        from django.contrib.auth import get_user_model
        User = get_user_model()
        if isinstance(self.user, User):
            return self.user.id
        else:
            return self.user

    def get_channel_name(self):
        return f"notif_user_{self.get_user_id()}"

    def save(self, *args, **kwargs):
        super(BaseNotificationModel, self).save(*args, **kwargs)
        self.dispatch()

    def __str__(self):
        return f"user_id:{self.get_user_id()}, title:{self.title}"


class BaseBroadcastNotificationModel(AbstractBaseNotificationModel):
    class NotificationTitle(models.TextChoices):
        system = 'system_notification'
        admin = 'admin_notification'

    title = models.CharField(choices=NotificationTitle.choices, max_length=50, default=NotificationTitle.system)
    user_read_at_receipts = models.JSONField(default=dict, null=True, blank=True)

    class Meta:
        abstract = True

    @classmethod
    def get_serializer_class(cls):
        from .serializers import BaseBroadcastNotificationSerializer
        return BaseBroadcastNotificationSerializer.get_subclass(cls.notification_class_name)

    def mark_as_read(self, user_id):
        if not self.user_read_at_receipts:
            self.user_read_at_receipts = {}

        if exists_user_read_at_receipt := self.user_read_at_receipts.get(user_id, None):
            print(f"notification already marked as read at receipt: {exists_user_read_at_receipt}")
            return False
        self.user_read_at_receipts[str(user_id)] = timezone.localtime()
        self.save(update_fields=['user_read_at_receipts'])
        print(f"notification marked as read for user {user_id} at {self.user_read_at_receipts.get(str(user_id))}")
        return True

    def get_read_at(self, user_id: str):
        if exists_user_read_at_receipt := self.user_read_at_receipts.get(user_id, None):
            return exists_user_read_at_receipt

        return None

    def save(self, *args, **kwargs):
        super(BaseBroadcastNotificationModel, self).save(*args, **kwargs)
        self.dispatch(broadcast=True)

    def get_channel_name(self):
        return "broadcast"

    def __str__(self):
        return f"broadcast message from user {self.requested_user}"



from rest_framework import serializers


class BaseSerializer(serializers.ModelSerializer):
    created_at = serializers.SerializerMethodField()
    modified_at = serializers.SerializerMethodField()

    def __init__(self, *args, **kwargs):
        fields = kwargs.pop('fields', None)
        excluded_fields = kwargs.pop('excluded_fields', None)
        super().__init__(*args, **kwargs)

        if fields is not None:
            # Drop any fields that are not specified in the `fields` argument
            allowed = set(fields)
            existing = set(self.fields.keys())
            for field_name in existing - allowed:
                self.fields.pop(field_name)

        if excluded_fields is not None:
            all_fields = set(self.fields.keys())
            disallowed = set(excluded_fields)
            for field_name in all_fields.intersection(disallowed):
                self.fields.pop(field_name)

    class Meta:
        model = None  # You need to set this in the derived serializers
        fields = '__all__'

    @staticmethod
    def get_created_at(obj):
        try:
            return obj.created_at.strftime("%Y-%m-%d %H:%M:%S")
        except Exception as err:
            return str(err)

    @staticmethod
    def get_modified_at(obj):
        try:
            return obj.modified_at.strftime("%Y-%m-%d %H:%M:%S")
        except Exception as err:
            return str(err)


class MessageResult:
    def __init__(self, response_text: str, success: bool = True, status: int = 200, **kwargs):
        self.response_text = response_text
        self.success = success
        self.status = status
        self.meta = kwargs

    def __str__(self):
        return self.response_text

    def __repr__(self):
        return self.response_text


def get_subclasses(_class, recursive=False):
    subclasses = _class.__subclasses__()
    if not recursive:
        return subclasses
    all_subclasses = subclasses.copy()
    for subclass in subclasses:
        all_subclasses.extend(subclass.get_all_subclasses())
    return all_subclasses



from itertools import chain

from pyasn1_modules.rfc2315 import data

from dj_notif.models import BaseNotificationModel, AbstractBaseNotificationModel, BaseBroadcastNotificationModel


class AbstractNotificationManager:
    def __init__(self):
        self.user = None

    @staticmethod
    def get_notification_models():
        return {model.notification_class_name: model for model in AbstractBaseNotificationModel.get_all_subclasses()}

    def get_notification_categories(self):
        return list(self.get_notification_models().keys())

    def get_notifications(self, broadcast=False):
        all_qs = []

        for cls in self.get_notification_models().values():
            if broadcast:
                qs = cls.objects.all()
            else:
                qs = cls.objects.filter(user=self.user)

            all_qs.append(qs)

        # Combine all querysets into one iterable
        combined = chain(*all_qs)

        # Sort by created_at (assuming it's in your BaseModel)
        sorted_notifications = sorted(combined, key=lambda n: n.created_at, reverse=True)

        return sorted_notifications


class NotificationManager(AbstractNotificationManager):
    def __init__(self, user):
        super().__init__()
        self.user = user

    @staticmethod
    def get_notification_models():
        return {model.notification_class_name: model for model in BaseNotificationModel.get_all_subclasses()}

    def get_categories_unread_count(self):
        notifications = self.get_categorized_notifications()
        return {notif_cls_name: notif_qs.filter(read_at=None).count() for
                notif_cls_name, notif_qs in notifications.items()}

    def get_notifications_serialized(self, unread_only=False):
        notifications_query = self.get_notifications()
        if unread_only:
            notifications_query = list(filter(lambda x: not x.read_at, notifications_query))
        return [x.__class__.get_serializer_class()(x).data for x in notifications_query]

    def get_categorized_notifications(self, unread_only=False):
        data = {}
        for cls_name, cls in self.get_notification_models().items():
            if not unread_only:
                data[cls_name] = cls.objects.filter(user=self.user)
            else:
                data[cls_name] = cls.objects.filter(user=self.user, read_at=None)
        return data

    def get_categorized_notifications_serialized(self, unread_only=False):
        user_notifications = self.get_categorized_notifications(unread_only=unread_only)
        serialized_notifications = {
            notif_cls_name: notif_qs.model.get_serializer_class()(
                notif_qs,
                many=True,
                excluded_fields=['category'],
                context={'unread_queryset': notif_qs.filter(read_at=None)}).data
            for notif_cls_name, notif_qs in user_notifications.items()}
        return serialized_notifications

    def get_category_notifications(self, category, unread_only=False):
        model = self.get_notification_models().get(category)
        if not unread_only:
            return model.objects.filter(user=self.user).prefetch_related('user')
        else:
            return model.objects.filter(user=self.user, read_at=None).prefetch_related('user')

    def get_category_notifications_serialized(self, category, unread_only=False):
        model = self.get_notification_models().get(category)
        notifications = self.get_category_notifications(category, unread_only=unread_only)
        return model.get_serializer_class()(notifications, many=True, excluded_fields=['category'],
                                            context={'unread_queryset': notifications.filter(read_at=None)}).data


class BroadcastNotificationManager(AbstractNotificationManager):
    @staticmethod
    def get_notification_models():
        return {model.notification_class_name: model for model in BaseBroadcastNotificationModel.get_all_subclasses()}

    def get_notifications_serialized(self, requested_user=None):
        notifications_query = self.get_notifications(broadcast=True)
        return [x.__class__.get_serializer_class()(x, context={'user': requested_user}).data for x in
                notifications_query]

    def get_categorized_notifications(self):
        data = {}
        for cls_name, cls in self.get_notification_models().items():
            data[cls_name] = cls.objects.all()
        return data

    def get_categorized_notifications_serialized(self, requested_user=None):
        user_notifications = self.get_categorized_notifications()

        serialized_notifications = {
            notif_cls_name: notif_qs.model.get_serializer_class()(
                notif_qs.order_by('-created_at'), many=True, excluded_fields=['category'],
                context={'user': requested_user}).data
            for notif_cls_name, notif_qs in user_notifications.items()}
        return serialized_notifications

    def get_category_notifications(self, category):
        model = self.get_notification_models().get(category)
        return model.objects.all()

    def get_category_notifications_serialized(self, category, requested_user=None):
        model = self.get_notification_models().get(category)
        serializer_class = model.get_serializer_class()
        notifications = model.objects.all().order_by('-created_at')

        return serializer_class(notifications, many=True, excluded_fields=['category'],
                                context={'user': requested_user}).data
